#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@created: 28.04.20
@author: felix
"""
import functools
import inspect
import pprint
import sys
import warnings
from functools import wraps
from typing import Type

from strongtyping._utils import _severity_level, action, remove_subclass
from strongtyping.cached_set import CachedSet
from strongtyping.config import SEVERITY_LEVEL
from strongtyping.strong_typing_utils import (
    TypeMisMatch,
    check_type,
    checking_typing_typedict_values,
    default_return_queue,
)


def match_typing(
    _func=None,
    *,
    excep_raise: Type[Exception] = TypeMisMatch,
    subclass: bool = False,
    severity="env",
    **kwargs,
):
    cached_enabled: int = kwargs.get("cache_size", 1)
    cached_set = CachedSet(cached_enabled) if cached_enabled > 0 else None

    def wrapper(func):
        # needed in py 3.10
        # globals().update(func.__globals__)

        arg_names = [name for name in inspect.signature(func).parameters]
        annotations = func.__annotations__
        severity_level = _severity_level(severity)

        @wraps(func)
        def inner(*args, **kwargs):
            if arg_names and severity_level > SEVERITY_LEVEL.DISABLED.value:

                args = remove_subclass(args, subclass)
                if cached_set is not None and func.__name__ not in ("__init__",):
                    # check if func with args and kwargs was checked once before with positive result
                    cached_key = (func, args.__str__(), kwargs.__str__())
                    if cached_key in cached_set:
                        return func(*args, **kwargs)

                # Thanks to Ruud van der Ham who find a better and more stable solution for check_args
                failed_params = tuple(
                    arg_name
                    for arg, arg_name in zip(args, arg_names)
                    if not check_type(arg, annotations.get(arg_name))
                )
                failed_params += tuple(
                    kwarg_name
                    for kwarg_name, kwarg in kwargs.items()
                    if not check_type(kwarg, annotations.get(kwarg_name))
                )

                if not default_return_queue.empty():
                    return default_return_queue.queue.pop()

                if failed_params:
                    annotated_values = {arg_name: arg for arg, arg_name in zip(args, arg_names)}
                    for kwarg_name, kwarg in kwargs.items():
                        annotated_values[kwarg_name] = kwarg

                    msg_list = "\nIncorrect parameter: ".join(
                        f"[{name}] `{pprint.pformat(annotated_values[name], width=20, depth=2)}`"
                        f"\n\trequired: {annotations[name]}"
                        for name in failed_params
                    )
                    msg = f"Incorrect parameter: {msg_list}"

                    if excep_raise is not None and severity_level == SEVERITY_LEVEL.ENABLED.value:
                        raise excep_raise(msg) from None
                    else:
                        warnings.warn(msg, RuntimeWarning)

                if cached_set is not None and func.__name__ not in ("__init__",):
                    cached_set.add(cached_key)
            return func(*args, **kwargs)

        inner.__fe_strng_mtch__ = 0
        return inner

    if _func is not None:
        return wrapper(_func)
    else:
        return wrapper


def add_required_methods_to_class(cls, inst):
    for method in ("__instancecheck__",):
        try:
            setattr(cls, method, getattr(inst, method))
        except AttributeError:
            continue


class MatchTypedDict:
    def __new__(cls, instance=None, *args, **kwargs):
        cls.cls = instance
        add_required_methods_to_class(cls, instance)
        return super().__new__(cls)

    def __init__(self, cls=None, *args, **kwargs):
        self.excep_raise = kwargs.pop("excep_raise", TypeMisMatch)
        self.cache_size = kwargs.pop("cache_size", 1)
        self.severity = kwargs.pop("severity", "env")
        self.cls = cls

    def __getattr__(self, item):
        return getattr(self.cls, item)

    @property
    def is_typed_dict(self):
        if hasattr(self.cls, "__orig_bases__"):
            return any(obj.__name__ == "TypedDict" for obj in self.cls.__orig_bases__)

    def create_error_msg(self, args: dict):
        return (
            f"Incorrect parameter: `{pprint.pformat(args, width=20, depth=2)}`"
            f"\n\trequired: {self.__annotations__}"
        )

    def __call__(self, *args, **kwargs):
        if self.is_typed_dict:
            arguments = kwargs if kwargs else args[0]
            if not checking_typing_typedict_values(arguments, self.__annotations__, self.__total__):
                raise self.excep_raise(self.create_error_msg(arguments))
        if self.cls:
            cls = self.cls(*args, **kwargs)
        else:
            cls = args[0]
        return cls


def match_class_typing(cls=None, **kwargs):
    excep_raise = kwargs.pop("excep_raise", TypeMisMatch)
    cache_size = kwargs.pop("cache_size", 1)
    severity = kwargs.pop("severity", "env")

    def __has_annotations__(obj):
        return hasattr(obj, "__annotations__")

    def __find_methods(_cls):
        return [
            func
            for func in dir(_cls)
            if callable(getattr(_cls, func))
            and __has_annotations__(getattr(_cls, func))
            and not hasattr(getattr(_cls, func), "__fe_strng_mtch__")
            and not isinstance(getattr(_cls, func), classmethod)
        ]

    def __add_decorator(_cls):
        severity_level = _severity_level(severity)
        if severity_level > SEVERITY_LEVEL.DISABLED.value:
            for method in __find_methods(_cls):
                try:
                    func = getattr(_cls, method)
                    is_static = "self" not in inspect.signature(func).parameters
                    setattr(
                        _cls,
                        method,
                        match_typing(
                            func,
                            severity=severity,
                            cache_size=cache_size,
                            excep_raise=excep_raise,
                            subclass=is_static,
                        ),
                    )
                except TypeError:
                    pass

    def wrapper(some_cls):
        def inner(*args, **cls_kwargs):
            __add_decorator(some_cls)
            return some_cls(*args, **cls_kwargs)

        inner._matches_class = True
        return inner

    if cls is not None:
        if sys.version_info.major >= 3 and sys.version_info.minor > 7:
            from typing import Type, _TypedDictMeta

            if isinstance(cls, _TypedDictMeta):
                return MatchTypedDict(cls)
        __add_decorator(cls)
        cls._matches_class = True
        return cls
    else:
        return wrapper


def getter(func):
    return action(func, "getter", match_typing)


def setter(func):
    return action(func, "setter", match_typing)


def getter_setter(func):
    return action(func, "getter_setter", match_typing)


class FinalClass:
    def __new__(cls, instance=None, *args, **kwargs):
        if args:
            raise RuntimeError(
                f"`class {instance}` can not inherit from `class {args[0][0].__name__}`"
            )
        cls.cls = instance
        return super().__new__(cls)

    def __init__(self, cls=None, *args, **kwargs):
        self.cls = cls

    def __getattr__(self, item):
        return getattr(self.cls, item)

    def __call__(self, *args, **kwargs):
        return self.cls(*args, **kwargs)

    def __repr__(self):
        return repr(self.cls)

    def __str__(self):
        return str(self.cls)

    @property
    def __doc__(self):
        return self.cls.__doc__



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@created: 13.07.20
@author: felix
"""
from collections import namedtuple
from keyword import iskeyword
from typing import Any, List, Tuple, Union

from strongtyping.docstring_typing import check_doc_str_type
from strongtyping.strong_typing import check_type, match_typing

use_match_typing = {True: check_type, False: check_doc_str_type}


@match_typing
def typed_namedtuple(
    typename: str,
    field_names: Union[List[str], str, List[Tuple[str, Any]]],
    *,
    rename: bool = False,
    defaults: Union[list, tuple] = None,
    module: str = None,
):
    # I could have just copied everything from namedtuple, but then I would have no learning effect
    """
    :param typename: the name of the new class same as in the original namedtuple
    :param field_names: as in the original field_names can be written as string or list
    to use typing write:
        'foo:int,bar:str,foobar:list
            or
        ['foo:int', 'bar:str', 'foobar:list']
            if no "typing" is inside of the field_names it will return the original namedtuple
    :param rename: same as in the original namedtuple
    :param defaults: the default values to use when instantiate, must match field_names length
    :param module: the module of the class same as in the original namedtuple
    :return:
    """

    def rename_fields():
        knowing = set()
        updated_dict = {}
        for index, name in enumerate(_field_types.keys()):
            if not name.isidentifier() or iskeyword(name) or name.startswith("_"):
                updated_dict[f"_{index}"] = _field_types[name]
            else:
                updated_dict[name] = _field_types[name]
            knowing.update(name)
        return updated_dict

    def validate_field_names():
        seen = set()
        for name in _field_types.keys():
            if name.startswith("_"):
                raise ValueError(f"Field names cannot start with an underscore: {name!r}")

            if not name.isidentifier():
                raise ValueError(
                    "Type names and field names must be valid " f"identifiers: {name!r}"
                )
            if iskeyword(name):
                raise ValueError("Type names and field names cannot be a " f"keyword: {name!r}")

            if name in seen:
                raise ValueError(f"Encountered duplicate field name: {name!r}")
            seen.add(name)

    def contains_typing(f_name: Union[str, tuple]) -> bool:
        return ":" in f_name or isinstance(f_name, tuple)

    def check_type(_value_dict: dict, use_mt: bool = False):
        failed_params = tuple(
            f"{k}: {v}"
            for k, v in _value_dict.items()
            if not use_match_typing[use_mt](v, _field_types[k])
        )
        if failed_params:
            msg = f"Incorrect parameters: {failed_params}"
            raise TypeError(msg)

    _fields = (
        [name.strip() for name in field_names.split(",")]
        if isinstance(field_names, (str,))
        else field_names
    )

    typing_true, typing_false = all(contains_typing(fn) for fn in _fields), not all(
        contains_typing(fn) for fn in _fields
    )

    if typing_false and not typing_true:
        if any(contains_typing(fn) for fn in _fields):
            raise TypeError("No mixing of typing and not typing supported")
        return namedtuple(typename, field_names, rename=rename, defaults=defaults, module=module)
    else:
        try:
            _field_types = {k: v for k, v in map(lambda x: x.split(":"), _fields)}
            _use_match = False
        except AttributeError:
            _field_types = {k: v for k, v in _fields}
            _use_match = True

        if rename is True:
            _field_types = rename_fields()
        else:
            validate_field_names()

        def _values_to_add(*args, **kwargs):
            _a = {k: v for k, v in zip(_field_types.keys(), args)}
            _b = {k: v for k, v in kwargs.items() if k in _field_types}
            return {**_a, **_b}

        def _values_with_defaults():
            if defaults is not None:
                if len(_field_types) != len(defaults):
                    raise TypeError("Default values must match with field names")
                _defaults = {k: v for k, v in zip(_field_types.keys(), defaults)}
                check_type(_defaults, _use_match)
                return _defaults
            else:
                raise TypeError(f"Initialise {typename} with values or add defaults")

        def __new__(cls, *args, **kwargs):
            _values = _values_to_add(*args, **kwargs)
            if not _values:
                _values = _values_with_defaults()
            if _values and defaults is not None:
                _values = _values_with_defaults()
                _values.update(**_values_to_add(*args, **kwargs))
            check_type(_values, _use_match)
            new_tuple = tuple.__new__(cls, _values.values())
            [setattr(new_tuple, k, v) for k, v in _values.items()]
            return new_tuple

        __new__.__doc__ = f"Create new instance of {typename}({_field_types.keys()})"

        if defaults is not None:
            __new__.__defaults__ = tuple(defaults)
            _field_defaults = _values_with_defaults()
        else:
            _field_defaults = None

        def _asdict(self):
            return {k: v for k, v in zip(_field_types.keys(), self)}

        def _replace(self, **kwargs):
            new_val = self._asdict()
            not_allowed = [k for k in kwargs.keys() if k not in self._field_types]
            if not_allowed:
                raise ValueError(f"Got unexpected field names: {not_allowed!r}")
            new_val.update(**kwargs)
            return self.__new__(self.__class__, **new_val)

        _replace.__doc__ = (
            f"Return a new {typename} object replacing specified fields with new values"
        )

        repr_fmt = "(" + ", ".join(f"{name}=%r" for name in _field_types.keys()) + ")"

        def __repr__(self):
            """Return a nicely formatted representation string"""
            return self.__class__.__name__ + repr_fmt % self

        def __getnewargs__(self):
            """Return self as a plain tuple.  Used by copy and pickle."""
            return tuple(self)

        class_doc = "\n".join([f":type {k}: {val}" for k, val in _field_types.items()])

        namespace = {
            "__new__": __new__,
            "__module__": module if module is not None else __name__,
            "_field_types": _field_types,
            "_asdict": _asdict,
            "_replace": _replace,
            "__getnewargs__": __getnewargs__,
            "__repr__": __repr__,
            "__doc__": f'{typename}({", ".join(_field_types.keys())})\n{class_doc}',
            "_field_defaults": _field_defaults,
        }

        if _use_match:
            namespace["__annotations__"] = _field_types
        return type(typename, (tuple,), namespace)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@created: 04.06.20
@author: felix
"""

__all__ = [
    "_utils",
    "strong_typing_utils",
    "strong_typing",
    "docstring_typing",
    "cached_set",
    "cached_dict",
    "type_namedtuple",
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@created: 12.05.21
@author: eisenmenger
"""
import inspect
import pprint
import re
import textwrap
from functools import wraps

from strongtyping.strong_typing_utils import get_origins, get_possible_types

Pattern = re.compile(r"(\$\d[a-zA-Z0-9, ()\n]+)")


def getsource(object):
    """Return the text of the source code for an object.

    The argument may be a module, class, method, function, traceback, frame,
    or code object.  The source code is returned as a single string.  An
    OSError is raised if the source code cannot be retrieved."""
    lines, lnum = inspect.getsourcelines(object)
    return "".join(lines[1:])


# def numpy_doc_string():
#     """
#     Some short description
#
#     Parameters
#     ----------
#     val_a : int
#         some infos if available
#     val_b : {int, float}, optional
#         default is 'value'
#
#     Returns
#     -------
#     string
#         some info if available
#
#     Raises
#     ------
#     TypeMisMatch
#         if the passed values not matching the required ones
#     """
#
#
# def rest_doc_string(val_a: int, val_b: int = 10) -> int:
#     """
#     Some short description
#
#     :param val_a: some info
#     :param val_b: (Default value = '10')
#     :type val_a: int
#     :returns: int
#     """


ARGUMENT_TYPE = {
    inspect.Parameter.KEYWORD_ONLY: "keyword only argument",
    inspect.Parameter.POSITIONAL_ONLY: "postional only argument",
    inspect.Parameter.POSITIONAL_OR_KEYWORD: "argument",
    inspect.Parameter.VAR_POSITIONAL: "variadic arguments",
    inspect.Parameter.VAR_KEYWORD: "variadic keyword arguments",
}


def union_types(val, type_origins):
    types = []
    for type_origin in get_possible_types(type_origins):
        try:
            types.append(type_origin.__name__)
        except AttributeError:
            types.append(get_type_info(val, type_origin))
    return " or ".join(types)


def get_type_info(val, type_origins):
    origins = get_origins(type_origins)
    val_origins = get_origins(val)
    if get_origins(val)[1] == "Union":
        try:
            return " or ".join([type_origin.__name__ for type_origin in type_origins])
        except AttributeError:
            return get_type_info(val, type_origins)
        except TypeError:
            return union_types(val, type_origins)
    elif origins[1] == "Union":
        return union_types(val, type_origins)
    elif origins[1].lower() == "dict":
        try:
            return f"{get_origins(val)[1]}({get_type_info(val, type_origins[0])}, {get_type_info(val, type_origins[1])})"
        except TypeError:
            text = ", ".join(
                [
                    get_type_info(val, type_origin)
                    for type_origin in get_possible_types(type_origins)
                ]
            )
            return f"{get_origins(val)[1]}({text})"
    elif origins[1].lower() in ("list", "tuple", "set"):
        text = ", ".join(
            [get_type_info(val, type_origin) for type_origin in get_possible_types(type_origins)]
        )
        if origins[1] != "None" and get_origins(val)[1] != origins[1]:
            return f"{origins[1]}({text})"
        return f"{get_origins(val)[1]}({text})"
    elif val_origins[1] == "Literal":
        text = " or ".join([f"`{elem}`" for elem in type_origins])
        return f"`{type(type_origins[0]).__name__}` allowed values are {text}"
    elif val_origins[1] == "TypedDict" or val_origins[1] == "_TypedDictMeta":
        required = " required" if val_origins[0].__total__ else ""
        fields = {
            key: get_type_info(key, val) for key, val in val_origins[0].__annotations__.items()
        }
        return f"{val.__name__}[TypedDict]{required} fields are \n\t`{pprint.pformat(fields, sort_dicts=False)}`"
    elif "Validator" in origins[1]:
        required_type, *not_needed = type_origins.__args__
        text = f"{get_type_info(val, required_type)}[{origins[1]}]"
        return text
    elif val_origins[1] == "args":
        return "tuple"
    elif val_origins[1] == "kwargs":
        return "dict"
    else:
        if type_origins:
            try:
                origins = ", ".join((type_origin.__name__ for type_origin in type_origins))
            except AttributeError:
                if len(type_origins) == 1:
                    return f"{get_origins(val)[1]}({get_type_info(val, type_origins[0])})"
                else:
                    return (
                        f"{get_origins(val)[1]}({get_type_info(val, type_origins[0])},"
                        f" {get_type_info(val, type_origins[1])})"
                    )
            except TypeError:
                try:
                    return type_origins.__name__
                except AttributeError:
                    return str(type_origins)
            else:
                return f"{get_origins(val)[1]}({origins})"
        else:
            if hasattr(val, "__name__"):
                return val.__name__
            return str(val)


def docs_from_typing_numpy_format(
    annotations, additional_infos, func_params, remove_linebreak, func_info
):
    doc_infos = ["Parameters", "----------"]
    type_infos = ["Returns", "-------"]
    func_param_keys = [
        key for key in func_params.keys() if key != "self" and key not in annotations
    ]
    annotation_keys = list(annotations.keys())

    for idx, key in enumerate((annotation_keys + func_param_keys), 1):
        if key != "return":
            val = annotations.get(key, func_params[key])
            type_origins = get_possible_types(val)
            predefined_info = "\n\t".join(additional_infos.get(f"${idx}", "").split("\n"))
            predefined_info = f"\n\t{predefined_info}" if predefined_info else ""

            info_str = f"{key} : {ARGUMENT_TYPE[func_params[key].kind]} of type {get_type_info(val, type_origins)}"

            if func_params[key].default is not inspect._empty:
                info_str = f"{info_str}\n\tDefault is {func_params[key].default}"

            info_str = f"{info_str}{predefined_info}"
            doc_infos.append(info_str)
    if "return" in annotations:
        doc_infos.append("")
        type_infos.append(get_type_info(annotations["return"], annotations["return"]))
    else:
        type_infos = []

    lb = "" if remove_linebreak else "\n"
    return lb + "\n".join(doc_infos + type_infos), func_info


def docs_from_typing_reST_format(
    annotations, additional_infos, func_params, remove_linebreak, func_info
):
    doc_infos = []
    type_infos = []
    func_param_keys = [
        key for key in func_params.keys() if key != "self" and key not in annotations
    ]
    annotation_keys = list(annotations.keys())
    for idx, key in enumerate((annotation_keys + func_param_keys), 1):
        if key != "return":
            val = annotations.get(key, func_params[key])
            type_origins = get_possible_types(val)
            predefined_info = "\n\t".join(additional_infos.get(f"${idx}", "").split("\n"))
            predefined_info = f"\n\t{predefined_info}" if predefined_info else ""
            info_str = f":param {key}: {ARGUMENT_TYPE[func_params[key].kind]} {predefined_info}"

            if func_params[key].default is not inspect._empty:
                info_str = f"{info_str} (Default value = {func_params[key].default})"

            doc_infos.append(info_str)

            type_infos.append(f":type {key}: {get_type_info(val, type_origins)}")

    if "return" in annotations:
        type_infos.append(
            f':returns: {get_type_info(annotations["return"], annotations["return"])}'
        )

    lb = "" if remove_linebreak else "\n"
    return lb + "\n".join(doc_infos + type_infos), func_info


def docs_from_typing(func, remove_linebreak, style):
    annotations = func.__annotations__
    func_params = inspect.signature(func).parameters
    if func.__doc__:
        if not func.__doc__.endswith("\n") and func.__doc__ != "":
            func_doc = f"{func.__doc__}\n"
        else:
            func_doc = func.__doc__
        additional_infos = Pattern.split(textwrap.dedent(func_doc))
        func_info = textwrap.dedent(additional_infos[0])
        additional_infos = dict(
            [
                tuple(filter(lambda x: len(x) > 1, map(str.strip, re.split(r"(\$\d)", info))))
                for info in additional_infos
                if info and info[0] == "$"
            ]
        )
    else:
        if func.__name__ != "__init__":
            func_info = f"Function {func.__name__}\n\n"
        else:
            func_info = "\n"
        additional_infos = {}
    if style == "rest":
        return docs_from_typing_reST_format(
            annotations, additional_infos, func_params, remove_linebreak, func_info
        )
    elif style == "numpy":
        return docs_from_typing_numpy_format(
            annotations, additional_infos, func_params, remove_linebreak, func_info
        )


def rest_docs_from_typing(_func=None, *, insert_at: str = None, remove_linebreak: bool = False):
    def wrapper(func):
        @wraps(func)
        def inner(*args, **kwargs):
            return func(*args, **kwargs)

        text, func_doc = docs_from_typing(func, remove_linebreak, style="rest")

        if insert_at is not None:
            inner.__doc__ = func_doc.replace(insert_at, text)
        else:
            inner.__doc__ = f"{func_doc}{text}"
        inner.has_auto_generated_docs = True
        return inner

    if _func is not None:
        return wrapper(_func)
    else:
        return wrapper


def numpy_docs_from_typing(_func=None, *, insert_at: str = None, remove_linebreak: bool = False):
    def wrapper(func):
        @wraps(func)
        def inner(*args, **kwargs):
            return func(*args, **kwargs)

        text, func_doc = docs_from_typing(func, remove_linebreak, style="numpy")

        if insert_at is not None:
            inner.__doc__ = func_doc.replace(insert_at, text)
        else:
            inner.__doc__ = f"{func_doc}{text}"
        inner.has_auto_generated_docs = True
        return inner

    if _func is not None:
        return wrapper(_func)
    else:
        return wrapper


def class_docs_from_typing(_cls=None, *, doc_type: str = "reST"):
    def wrapper(cls):
        docs_formatter = (
            rest_docs_from_typing if doc_type.lower() == "rest" else numpy_docs_from_typing
        )
        cls.__doc__ = f"{cls.__doc__}{docs_formatter(cls.__init__).__doc__}"
        cls.__init__.__doc__ = ""
        users_funcs = [
            func
            for func in dir(cls)
            if inspect.isfunction(getattr(cls, func)) and func != "__init__"
        ]
        for func in users_funcs:
            cls_method = getattr(cls, func)
            if cls_method.__annotations__ and not hasattr(cls_method, "has_auto_generated_docs"):
                cls_method.__doc__ = docs_formatter(getattr(cls, func)).__doc__
        return cls

    if _cls is not None:
        return wrapper(_cls)
    else:
        return wrapper



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@created: 09.07.20
@author: felix
"""
from functools import partial

from strongtyping.strong_typing import match_typing


def action(f, frefs):

    """
    This code is original from Ruud van der Ham https://github.com/salabim/easy_property
    """
    if f.__qualname__ == action.qualname:
        if any(action.f[fref] is not None for fref in frefs.split("_")):
            raise AttributeError(f"decorator defined twice")
    else:
        action.f.update({}.fromkeys(action.f, None))  # reset all values to None
        action.qualname = f.__qualname__
    action.f.update({}.fromkeys(frefs.split("_"), f))  # set all frefs values to f

    # this line was added by myself
    action.f["setter"] = (
        match_typing(action.f["setter"]) if action.f["setter"] is not None else None
    )

    return property(
        *(
            action.f[ref] if (ref != "documenter" or action.f[ref] is None) else action.f[ref](0)
            for ref in action.f
        )
    )


action.qualname = None
action.f = dict.fromkeys(["getter", "setter", "deleter", "documenter"], None)

globals().update(
    {fref: partial(action, frefs=fref) for fref in {**action.f, "getter_setter": None}}
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@created: 30.05.21
@author: felix
"""
import inspect
import json
import weakref
from functools import partial
from typing import Any, _GenericAlias, _SpecialForm, _type_repr


class _Validator(_GenericAlias, _root=True):  # type: ignore
    _name = "Validator"

    def __getitem__(self, item):
        return super().__getitem__(item)

    def __hash__(self):
        if len(self.__args__) > 2:
            return hash(frozenset([self.__args__[:-1], json.dumps(self.__args__[-1])]))
        return hash(frozenset(self.__args__))

    def __repr__(self):
        args = self.__args__
        validator = args[1].func if isinstance(args[1], partial) else args[1]
        func_name = validator.__name__
        validation_function_file = inspect.getfile(validator)
        validation_body, validation_line = inspect.getsourcelines(validator)
        validation_lines = validation_line + len(validation_body)
        func_info = (
            f"{func_name}(<{validation_function_file}>, "
            f"lines={validation_line}-{validation_lines})"
        )
        return f"strong_typing_utils.Validator[{_type_repr(args[0])}, {func_info}]"


class _IterValidator(_GenericAlias, _root=True):  # type: ignore
    _name = "IterValidator"

    def __getitem__(self, item):
        return super().__getitem__(item)

    def __hash__(self):
        if len(self.__args__) > 2:
            return hash(frozenset([self.__args__[:-1], json.dumps(self.__args__[-1])]))
        return hash(frozenset(self.__args__))

    def __repr__(self):
        args = self.__args__
        validator = args[1].func if isinstance(args[1], partial) else args[1]
        func_name = validator.__name__
        validation_function_file = inspect.getfile(validator)
        validation_body, validation_line = inspect.getsourcelines(validator)
        validation_lines = validation_line + len(validation_body)
        func_info = (
            f"{func_name}(<{validation_function_file}>, "
            f"lines={validation_line}-{validation_lines})"
        )
        return f"strong_typing_utils.IterValidator[{_type_repr(args[0])}, {func_info}]"


@_SpecialForm  # type: ignore
def Validator(self, parameters, *args, **kwargs):
    if isinstance(parameters, _GenericAlias):
        raise TypeError("Validator needs min 2 values. Validator[type, function]")
    if not parameters:
        raise TypeError("Cannot take a Validator of no type/function.")
    if len(parameters) > 3:
        raise TypeError("Validator takes max 3 values.")
    if not inspect.isfunction(parameters[1]) and not isinstance(parameters[1], partial):
        raise TypeError("Validator[..., arg]: arg should be a function.")
    return _Validator(self, parameters)


@_SpecialForm  # type: ignore
def IterValidator(self, parameters, *args, **kwargs):
    if isinstance(parameters, _GenericAlias):
        raise TypeError("Validator needs min 2 values. Validator[type, function]")
    if not parameters:
        raise TypeError("Cannot take a Validator of no type/function.")
    if len(parameters) > 3:
        raise TypeError("Validator takes max 3 values.")
    if not inspect.isfunction(parameters[1]) and not isinstance(parameters[1], partial):
        raise TypeError("Validator[..., arg]: arg should be a function.")
    return _IterValidator(self, parameters)


class FrozenType:
    __slots__ = ("required_type", "stored_value", "weakref")

    def __init__(self, required_type, stored_value=None):
        self.weakref = weakref.WeakKeyDictionary()
        self.required_type = required_type
        self.stored_value = stored_value

    def __get__(self, instance=None, owner=None):
        return self.weakref.get(instance, (None, self.stored_value))[1]

    def __set__(self, instance=None, value=None):
        if isinstance(value, tuple) and isinstance(value[0], FrozenType):
            required_type, stored_value = self.weakref.get(
                instance, (self.required_type, self.stored_value)
            )
            new_required_type, original_type = value[1], required_type
            try:
                updated_stored_value = new_required_type(stored_value)
            except TypeError:
                raise TypeError(f"Cannot cast {self.required_type} to {value[1]}")
            else:
                self.required_type = new_required_type
                self.stored_value = updated_stored_value
                self.weakref[value[-1]] = (new_required_type, updated_stored_value)
        else:
            if not isinstance(value, self.weakref.get(instance, (self.required_type,))[0]):
                self.error_msg(value)
            else:
                self.weakref[instance] = (self.required_type, value)

    def __repr__(self):
        return repr(self.required_type)

    def __str__(self):
        return str(self.stored_value)

    def __doc__(self):
        return self.required_type.__doc__

    @classmethod
    def cast(cls, instance, origin, new):
        """
        change the type explicit not implicit by accident
        """
        try:
            new(origin)
        except TypeError:
            raise TypeError(f"Cannot cast {type(origin)} to {new}")
        return FrozenType(FrozenType), new, instance

    def error_msg(self, value: Any, attribute_name: str = "This") -> Exception:
        raise TypeError(
            f"`{attribute_name}` is a final type. "
            f"\n\tYou cannot assign {type(value)} to {self.required_type}"
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@created: 11.06.20
@author: felix
"""
import builtins
import functools
import inspect
import re
import typing
import warnings
from types import FunctionType, MethodType

from strongtyping._utils import _get_new, _severity_level, action, remove_subclass
from strongtyping.cached_set import CachedSet
from strongtyping.strong_typing import TypeMisMatch

TYPE_EXTRACTION_PATTERN = r"(^[:a-zA-Z0-9 _-]+(:))"
PATTERN_1 = r""
EXTRACT_PARAM_NAME_PATTERN = r"(?::(param|parameter|arg|argument|key|keyword|type|vartype)\W)"
TUPLE_PATTERN = r"\(([^)]+)\)"
PATTERN = r"[\(\)\[\[]"
OR_PATTERN = r"([\(\[]\w+)\W+(or)\W+(\w+)"
COMMA_PATTERN = r"([\(\[]\w+)\W+(\w+)"
REMOVE_PATTERN = r"[\(\[\)\]]"
FM_PATTERN = r"([FM][a-z]{5,7}Type)"


def separate_param_type(docstring_type_part: str) -> tuple:
    t = re.split(TYPE_EXTRACTION_PATTERN, docstring_type_part)
    clean = [re.sub(PATTERN_1, "", part.strip()) for part in t if part]
    return re.sub(EXTRACT_PARAM_NAME_PATTERN, "", clean[0]).replace(":", ""), clean[-1]


possible_types = {"(": tuple, "[": list, "{": set}


def param_attr(attr: str):
    """
    :return: builtin class or typing instance
    """
    try:
        return getattr(typing, attr)
    except AttributeError:
        return getattr(builtins, attr)


def get_container_types(ttype_of: str) -> typing.Union[None, tuple]:
    pattern = r"([\(\[]\w+)" if "or" not in ttype_of else OR_PATTERN
    pattern = pattern if ", " not in ttype_of else COMMA_PATTERN
    sub_pattern = re.findall(pattern, ttype_of)
    sub_pattern = sub_pattern if "or" not in ttype_of else sub_pattern[0]
    sub_pattern = sub_pattern if ", " not in ttype_of else sub_pattern[0]
    try:
        container_types = tuple(
            param_attr(re.sub(REMOVE_PATTERN, "", t).strip()) for t in sub_pattern if t != "or"
        )
    except TypeError:
        container_types = None
    return container_types


def get_or_types(ttype: str) -> list:
    if " or " in ttype:
        return [t for t in re.findall(r"(\w+)\W+(or)\W+(\w+)", ttype)[0] if t != "or"]
    else:
        return re.findall(r"\w{2,}", ttype)


def is_tuple(arg, type_of: str):
    container_types = get_container_types(type_of)
    sub_types = (
        all(isinstance(a, container_types) for a in arg) and len(arg) == len(container_types)
        if container_types
        else True
    )
    return isinstance(arg, tuple) and sub_types


def is_list(arg, type_of: str):
    container_types = get_container_types(type_of)
    sub_types = all(isinstance(a, container_types) for a in arg) if container_types else True
    return isinstance(arg, list) and sub_types


def is_dict(arg, type_of: str):
    container_types = get_container_types(type_of)
    sub_types = (
        all(isinstance(k, container_types[0]) for k in arg.keys())
        and all(isinstance(v, container_types[1]) for v in arg.values())
        if container_types
        else True
    )
    return isinstance(arg, dict) and sub_types


def is_set(arg, type_of: str):
    return isinstance(arg, set)


def is_function_or_method_type(arg, type_of):
    type_dict = {"F": isinstance(arg, FunctionType), "M": isinstance(arg, MethodType)}
    return type_dict[type_of[0]]


options = {"tuple": is_tuple, "list": is_list, "set": is_set, "dict": is_dict}


def check_doc_str_type(arg, type_of):
    check_result = True
    if type_of is not None:
        try:
            return options[re.split(REMOVE_PATTERN, type_of)[0]](arg, type_of)
        except KeyError:
            if len(re.findall(FM_PATTERN, type_of)) == 1:
                return is_function_or_method_type(arg, type_of)
            try:
                return isinstance(arg, tuple(map(param_attr, get_or_types(type_of))))
            except AttributeError:
                return arg.__class__.__name__ == type_of
    return check_result


def is_type_info(docstring_line: str) -> bool:
    allowed = [":type", ":vartype"]
    return any(docstring_line.startswith(a) for a in allowed)


def is_param_info(docstring_line: str) -> bool:
    allowed = [":param", ":parameter", ":arg", ":argument", ":key", ":keyword"]
    return any(docstring_line.startswith(a) for a in allowed)


def extract_docstring_param_types(func) -> dict:
    """
    extract the types to the defined parameters from the docstring
    """
    param = [
        separate_param_type(string)[0].split()
        for string in inspect.getdoc(func).split("\n")
        if is_param_info(string)
    ]
    docstring = [
        separate_param_type(string)
        for string in inspect.getdoc(func).split("\n")
        if is_type_info(string)
    ]
    docstring += [list(reversed(p)) for p in param if len(p) > 1]
    _docstring_types = {ds[0]: ds[1] for ds in docstring}
    # there is mismatch when user will mix type and param to bring them in the right order
    # we will look at the signature and recreate the previous dict to the final one
    return {k: _docstring_types.get(k, k) for k in inspect.signature(func).parameters.keys()}


def match_docstring(
    _func=None,
    *,
    excep_raise: Exception = TypeMisMatch,
    cache_size=0,
    subclass: bool = False,
    severity="env",
    **kwargs,
):
    cached_set = None if cache_size == 0 else CachedSet(cache_size)

    def wrapper(func):
        docstring_types = extract_docstring_param_types(func)

        severity_level = _severity_level(severity)

        @functools.wraps(func)
        def inner(*args, **kwargs):
            if severity_level > 0:

                args = remove_subclass(args, subclass)

                if cached_set is not None:
                    # check if func with args and kwargs was checked once before with positive result
                    cached_key = (func, args.__str__(), kwargs.__str__())
                    if cached_key in cached_set:
                        return func(*args, **kwargs)

                if "self" in docstring_types:
                    docstring_types["self"] = args[0].__class__.__name__
                if "cls" in docstring_types:
                    docstring_types["cls"] = args[0].__name__
                # Thanks to Ruud van der Ham who find a better and more stable solution for check_args
                failed_params = tuple(
                    arg_name
                    for arg, arg_name in zip(args, docstring_types)
                    if not check_doc_str_type(arg, docstring_types.get(arg_name))
                )
                failed_params += tuple(
                    kwarg_name
                    for kwarg_name, kwarg in kwargs.items()
                    if not check_doc_str_type(kwarg, docstring_types.get(kwarg_name))
                )
                if failed_params:
                    msg = f'Incorrect parameters: {", ".join(f"{name}: {docstring_types[name]}" for name in failed_params)}'
                    if excep_raise is not None and severity_level == 1:
                        raise excep_raise(msg)
                    else:
                        warnings.warn(msg, RuntimeWarning)

                if cached_set is not None:
                    cached_set.add(cached_key)

            return func(*args, **kwargs)

        inner.__fe_strng_mtch__ = 0
        return inner

    if _func is not None:
        return wrapper(_func)
    else:
        return wrapper


def match_class_docstring(
    _cls=None,
    *,
    excep_raise: Exception = TypeError,
    cache_size=0,
    severity="env",
    **kwargs,
):
    def wrapper(cls):

        severity_level = _severity_level(severity)

        def inner(*args, **kwargs):
            if severity_level > 0:
                cls.__new__ = _get_new(match_docstring, excep_raise, cache_size, severity, **kwargs)
                if hasattr(cls.__init__, "__annotations__"):
                    cls.__init__ = match_docstring(cls.__init__)
            return cls(*args, **kwargs)

        return inner

    if _cls is not None:
        return wrapper(_cls)
    else:
        return wrapper


def getter(func):
    return action(func, "getter", match_docstring)


def setter(func):
    return action(func, "setter", match_docstring)


def getter_setter(func):
    return action(func, "getter_setter", match_docstring)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@created: 20.07.20
@author: felix
"""
from enum import Enum
from os import environ


class SEVERITY_LEVEL(Enum):
    DISABLED = 0
    ENABLED = 1
    WARNING = 2

    @property
    def value_as_str(self):
        return str(self.value)


def set_severity_level(_level: SEVERITY_LEVEL):
    environ["ST_SEVERITY"] = _level.value_as_str



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@created: 19.11.20
@author: felix
"""
import inspect
import os
import typing
from functools import lru_cache, partial
from queue import Queue
from typing import Any, TypeVar, _GenericAlias, _SpecialForm, _type_repr  # type: ignore

from strongtyping._utils import install_st_m

install_st_m()

try:
    from strongtyping_modules.strongtyping_modules import dict_elements  # type: ignore
    from strongtyping_modules.strongtyping_modules import list_elements  # type: ignore
    from strongtyping_modules.strongtyping_modules import set_elements  # type: ignore
    from strongtyping_modules.strongtyping_modules import tuple_elements  # type: ignore
except ImportError as e:
    extension_module: bool = False
else:
    extension_module = bool(int(os.environ["ST_MODULES_INSTALLED"]))


empty = object()
default_return_queue = Queue()


class TypeMisMatch(AttributeError):
    def __init__(self, message):
        super().__init__()
        print(message)


class ValidationError(Exception):
    def __init__(self, message):
        super().__init__()
        print(message)


typing_base_class = typing._GenericAlias  # type: ignore


@lru_cache(maxsize=1024)
def get_possible_types(typ_to_check, origin_name: str = "") -> typing.Union[tuple, None]:
    """
    :param typ_to_check: some typing like List[str], Dict[str, int], Tuple[Union[str, int], List[int]]
    :param origin_name: the name of the origin
    :return: the inner types, classes of the type
        - List[str] = (str, )
        - Dict[str, int] = (str, int, )
        - Tuple[Union[str, int], List[int]] = (Union[str, int], List[int], )
    """
    if origin_name == "typeddict":
        # we can ensure now that we use a python version which has typing.TypedDict
        return typ_to_check

    if extension_module:
        if not hasattr(typ_to_check, "__args__"):
            try:
                return typ_to_check.__origin__
            except AttributeError:
                return None
    if hasattr(typ_to_check, "__args__") and typ_to_check.__args__ is not None:
        return tuple(typ for typ in typ_to_check.__args__ if not isinstance(typ, TypeVar))
    else:
        return None


def get_origins(typ_to_check: Any) -> tuple:
    """
    :param typ_to_check: typ_to_check: some typing like List[str], Dict[str, int], Tuple[Union[str, int], List[int]]
    :return: the class, alias_class and the class name
        - List[str] = (list, 'List')
        - Dict[str, int] = (dict, 'Dict')
        - Tuple[Union[str, int], List[int]] = (tuple, 'Tuple)
        - FunctionType = (None, 'None')
    """
    origin, origin_name = None, ""

    if hasattr(typ_to_check, "__annotations__") and hasattr(typ_to_check, "__orig_bases__"):
        return typ_to_check, typ_to_check.__orig_bases__[0].__name__

    if typing.is_typeddict(typ_to_check):
        return typ_to_check, typ_to_check.__class__.__name__

    if hasattr(typ_to_check, "__origin__") or hasattr(typ_to_check, "__orig_bases__"):
        if hasattr(typ_to_check, "__origin__") and hasattr(typ_to_check.__origin__, "__name__"):
            origin = origin_name = typ_to_check.__origin__.__name__
        elif hasattr(typ_to_check, "__origin__") or hasattr(typ_to_check, "__orig_bases__"):
            if hasattr(typ_to_check, "__origin__") and typ_to_check.__origin__ is not None:
                origin = typ_to_check.__origin__
            else:
                origin = typ_to_check.__orig_bases__

    if hasattr(origin, "_name"):
        origin_name = origin._name  # type: ignore
    else:
        if hasattr(typ_to_check, "_name"):
            origin_name = typ_to_check._name if typ_to_check._name is not None else origin_name
    return origin, origin_name


def checking_typing_dict(arg: Any, possible_types: tuple, *args):
    if not isinstance(arg, dict):
        return False
    if isinstance(arg, dict) and not possible_types:
        return True
    try:
        key, val = possible_types
    except (ValueError, TypeError):
        return isinstance(arg, dict)
    else:
        try:
            result_key = all(check_type(a, key) for a in arg.keys())
        except AttributeError:
            result_key = all(isinstance(k, key) for k in arg.keys())
        try:
            result_val = all(check_type(a, val) for a in arg.values())
        except AttributeError:
            result_val = all(isinstance(v, val) for v in arg.values())
        return result_key and result_val


def checking_typing_set(arg: Any, possible_types: tuple, *args, **kwargs):
    if not possible_types:
        return isinstance(arg, set)
    possible_type = possible_types[0]
    return isinstance(arg, set) and all(
        check_type(argument, possible_type, **kwargs) for argument in arg
    )


def checking_typing_type(arg: Any, possible_types: tuple, *args, **kwargs):
    try:
        arguments = arg.__mro__
    except AttributeError:
        return any(
            check_type(arg, possible_type, mro=False, **kwargs) for possible_type in possible_types
        )
    else:
        return any(
            check_type(arguments, possible_type, mro=True, **kwargs)
            for possible_type in possible_types
        )


def checking_typing_union(arg: Any, possible_types: tuple, mro, **kwargs):
    if mro:
        return any(pssble_type in arg for pssble_type in possible_types)
    try:
        is_instance = isinstance(arg, possible_types)
    except TypeError:
        return any(check_type(arg, typ, **kwargs) for typ in possible_types)
    else:
        if not is_instance:
            return False
        else:
            return validate_object(arg, kwargs.get("validation_with"))


def checking_typing_optional(arg: Any, possible_types: tuple, mro, **kwargs):
    return arg is None or check_type(arg, possible_types[0])


def checking_typing_iterator(arg: Any, *args, **kwargs):
    return hasattr(arg, "__iter__") and hasattr(arg, "__next__")


def checking_typing_callable(arg: Any, possible_types: tuple, *args, **kwargs):
    insp = inspect.signature(arg)
    return_val = insp.return_annotation == possible_types[-1]
    params = insp.parameters
    return return_val and all(p.annotation == pt for p, pt in zip(params.values(), possible_types))


def checking_typing_tuple(arg: Any, possible_types: tuple, *args, **kwargs):
    if not possible_types:
        return isinstance(arg, tuple)
    if Ellipsis in possible_types and isinstance(arg, tuple):
        if not arg:
            return True
        return checking_ellipsis(arg, possible_types, **kwargs)
    if not isinstance(arg, tuple) or not (len(arg) == len(possible_types)):
        return False
    return all(check_type(argument, typ, **kwargs) for argument, typ in zip(arg, possible_types))


def checking_typing_list(arg: Any, possible_types: tuple, *args, **kwargs):
    if not isinstance(arg, list):
        return False
    if isinstance(arg, list) and not possible_types:
        return True
    possible_type = possible_types[0]
    return all(check_type(argument, possible_type, **kwargs) for argument in arg)


def checking_ellipsis(arg, possible_types, *args, **kwargs):
    possible_types = [pt for pt in possible_types if pt is not Ellipsis]
    possible_type = possible_types[0]
    return all(check_type(argument, possible_type, **kwargs) for argument in arg)


def checking_typing_json(arg, possible_types, *args, **kwargs):
    try:
        possible_types.dumps(arg)
    except TypeError:
        return isinstance(arg, str)
    else:
        return True


def checking_typing_generator(arg, possible_types, *args, **kwargs):
    return hasattr(arg, "send") and hasattr(arg, "throw") and hasattr(arg, "__next__")


def checking_typing_literal(arg, possible_types, *args, **kwargs):
    return arg in possible_types


def checking_typing_validator(arg, possible_types, *args, **kwargs):
    if len(possible_types) == 2:
        default_return = empty
        required_type, validation = possible_types
    else:
        required_type, validation, default_return = possible_types
    if validation(arg) is False:
        if default_return is not empty:
            default_return_queue.put(default_return)
            return None
        if isinstance(validation, partial):
            validation = validation.func
        validation_function_file = inspect.getfile(validation)
        validation_body, validation_line = inspect.getsourcelines(validation)
        raise ValidationError(
            f"Argument: `{arg}` did not pass the validation defined here "
            f'\n\tFile: "{validation_function_file}", line {validation_line}'
            f"\n\tName: {validation.__name__}"
        )

    try:
        return isinstance(arg, required_type)
    except TypeError:
        return check_type(arg, required_type, **kwargs)


def checking_typing_itervalidator(arg, possible_types, *args, **kwargs):
    required_type, validation = possible_types
    return check_type(arg, required_type, validation_with=validation, **kwargs)


def checking_typing_iterable(arg: Any, possible_types: tuple, *args, **kwargs):
    if not hasattr(arg, "__iter__"):
        return False
    pssble_type = possible_types[0]
    return all(check_type(argument, pssble_type, **kwargs) for argument in arg)


def checking_typing_typedict_values(args: dict, required_types: dict, total: bool):
    if total:
        return all(check_type(args[key], val) for key, val in required_types.items())
    fields_to_check = {key: val for key, val in required_types.items() if key in args}
    return all(check_type(args[key], val) for key, val in fields_to_check.items())


def checking_typing_class(arg: Any, possible_types: tuple, *args, **kwargs):
    return isinstance(arg, possible_types)


def checking_typing_typeddict(arg: Any, possible_types: Any, *args, **kwargs):
    total = possible_types.__total__
    required_fields = possible_types.__annotations__
    if total:
        if not all(field in arg for field in required_fields):
            return False
    return checking_typing_typedict_values(arg, required_fields, total)


def module_checking_typing_list(arg: Any, possible_types: Any):
    if (
        not hasattr(possible_types, "__args__")
        or not possible_types.__args__
        or all(isinstance(pt, TypeVar) for pt in possible_types.__args__)
    ):
        return isinstance(arg, list)
    return bool(list_elements(arg, possible_types))


def module_checking_typing_dict(arg: Any, possible_types: Any):
    if (
        not hasattr(possible_types, "__args__")
        or not possible_types.__args__
        or all(isinstance(pt, TypeVar) for pt in possible_types.__args__)
    ):
        return isinstance(arg, dict)
    return bool(dict_elements(arg, possible_types))


def module_checking_typing_set(arg: Any, possible_types: Any):
    if (
        not hasattr(possible_types, "__args__")
        or isinstance(possible_types.__args__[0], TypeVar)
        or all(isinstance(pt, TypeVar) for pt in possible_types.__args__)
    ):
        return isinstance(arg, set)
    return bool(set_elements(arg, possible_types))


def module_checking_typing_tuple(arg: Any, possible_types: Any):
    if (
        not hasattr(possible_types, "__args__")
        or not possible_types.__args__
        or all(isinstance(pt, TypeVar) for pt in possible_types.__args__)
    ):
        return isinstance(arg, tuple)
    return bool(tuple_elements(arg, possible_types))


def module_checking_typing_validator(arg, possible_types, *args, **kwargs):
    try:
        required_type, validation = possible_types.__args__
    except ValueError:
        required_type, validation, _ = possible_types.__args__
    if validation(arg) is False:
        if isinstance(validation, partial):
            validation = validation.func
        validation_function_file = inspect.getfile(validation)
        validation_body, validation_line = inspect.getsourcelines(validation)
        raise ValidationError(
            f"Argument: `{arg}` did not pass the validation defined here "
            f'\n\tFile: "{validation_function_file}", line {validation_line}'
            f"\n\tName: {validation.__name__}"
        )
    return check_type(arg, required_type, **kwargs)


def validate_object(value, validation_func=None):
    if validation_func:
        return validation_func(value)
    return True


supported_typings = vars()
if extension_module:
    m = [f"module_checking_typing_{t}" for t in ("list", "dict", "set", "tuple", "validator")]
    supported_modules = {k: v for k, v in vars().items() if k in m}
else:
    supported_modules = {}


def check_type(argument, type_of, mro=False, **kwargs):
    # if int(py_version) >= 10 and isinstance(type_of, (str, bytes)):
    #     type_of = eval(type_of, locals(), globals())
    if checking_typing_generator(argument, type_of):
        # generator will be exhausted when we check it, so we return it without any checking
        return argument

    check_result = True
    if type_of is not None:
        origin, origin_name = get_origins(type_of)
        origin_name = origin_name.lower()

        if "any" in origin_name:
            return validate_object(argument, kwargs.get("validation_with"))
        if "json" in origin_name or "json" in str(type_of):
            return supported_typings["checking_typing_json"](argument, type_of, mro)

        try:
            return supported_modules[f"module_checking_typing_{origin_name}"](argument, type_of)
        except KeyError:
            pass

        if isinstance(type_of, typing_base_class) or origin is not None:
            try:
                return supported_typings[f"checking_typing_{origin_name}"](
                    argument, get_possible_types(type_of, origin_name), mro, **kwargs
                )
            except AttributeError:
                return isinstance(argument, type_of.__args__)
        elif isinstance(type_of, str):
            return argument.__class__.__name__ == type_of
        elif origin_name in ("_typeddictmeta", "matchtypeddict"):
            return checking_typing_typeddict(argument, get_possible_types(type_of, "typeddict"))
        elif mro:
            if origin_name == "union":
                possible_types = get_possible_types(type_of)
                return supported_typings[f"checking_typing_{origin_name}"](
                    argument, possible_types, mro
                )
            return type_of in argument
        else:
            try:
                is_instance = isinstance(argument, type_of)
            except TypeError:
                return isinstance(argument, type_of._subs_tree()[1:])
            else:
                if not is_instance:
                    return False
                return validate_object(argument, kwargs.get("validation_with"))
    return check_result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@created: 20.07.20
@author: felix
"""
import inspect
import logging
import os
from types import MethodType
from typing import Any, Type, Union

from strongtyping.config import SEVERITY_LEVEL

logger = logging.getLogger(__name__)


def remove_subclass(args, subclass):
    if len(args) == 1:
        return args
    cls = args[0] if subclass else None
    if cls is not None:
        args = args[1:]
    return args


SEVERITY_CONFIG = {
    "warning": SEVERITY_LEVEL.WARNING,
    "disable": SEVERITY_LEVEL.DISABLED,
    "enable": SEVERITY_LEVEL.ENABLED,
}


def _severity_level(severity_: Union[str, SEVERITY_LEVEL]):
    if severity_ == "env":
        _level = os.environ.get("ST_SEVERITY", "1")
        try:
            return int(_level)
        except (TypeError, ValueError):
            return SEVERITY_CONFIG[_level].value
    else:
        return severity_.value  # type: ignore


exclude_builtins = dir(object)


def _get_new(
    typing_func, excep_raise: Type[Exception] = TypeError, cache_size=0, severity="env", **kwargs
):
    def new_with_match_typing(cls_, *args, **kwargs):
        def add_match_typing(obj: object, attr: str) -> bool:
            if (
                hasattr(getattr(cls_, attr), "__annotations__")
                and getattr(cls_, attr).__class__.__name__ != "property"
                and not hasattr(getattr(x, attr), "__fe_strng_mtch__")
            ):
                type_annotations = getattr(getattr(cls_, attr), "__annotations__")
                return len([i for i in type_annotations.keys() if i != "return"]) > 0
            return False

        x = object.__new__(cls_)
        [
            setattr(
                x,
                cls_func,
                MethodType(
                    typing_func(
                        getattr(x, cls_func),
                        excep_raise=excep_raise,
                        cache_size=cache_size,
                        subclass=True,
                        severity=severity,
                    ),
                    x,
                ),
            )
            for cls_func in dir(x)
            if cls_func not in exclude_builtins and add_match_typing(x, cls_func)
        ]
        return x

    return new_with_match_typing


def install_st_m():
    import os

    try:
        from strongtyping_modules.install import install  # type: ignore
    except ImportError:
        os.environ["ST_MODULES_INSTALLED"] = "0"
    else:
        if not bool(int(os.environ.get("ST_MODULES_INSTALLED", "0"))):
            logger.info("strongtyping_modules will be installed")
            install()
            os.environ["ST_MODULES_INSTALLED"] = "1"


def action(f, frefs, type_function):  # type: ignore
    """
    This code is original from Ruud van der Ham https://github.com/salabim/easy_property
    """
    if f.__qualname__ == action.qualname:
        if any(action.f[fref] is not None for fref in frefs.split("_")):
            raise AttributeError(f"decorator defined twice")
    else:
        action.f.update({}.fromkeys(action.f, None))  # reset all values to None
        action.qualname = f.__qualname__
    action.f.update({}.fromkeys(frefs.split("_"), f))  # set all frefs values to f

    # this line was added by myself
    action.f["setter"] = (
        type_function(action.f["setter"]) if action.f["setter"] is not None else None
    )

    return property(
        *(
            action.f[ref] if (ref != "documenter" or action.f[ref] is None) else action.f[ref](0)
            for ref in action.f
        )
    )


action.qualname = None  # type: ignore
action.f = dict.fromkeys(["getter", "setter", "deleter", "documenter"], None)  # type: ignore
